import asynctest
from unittest.mock import MagicMock


class ResponseMock():

    json = asynctest.CoroutineMock(return_value={})
    raise_for_status = asynctest.CoroutineMock()
    status = 200


class SessionMock(MagicMock):

    request = asynctest.CoroutineMock(return_value=ResponseMock())
    close = asynctest.CoroutineMock()




class AIONetboxException(Exception):
    pass


class MissingRequiredParam(AIONetboxException):
    pass


class BadRequest(AIONetboxException):
    pass


class InvalidResponse(AIONetboxException):
    pass


class InvalidRequest(AIONetboxException):
    pass



from aionetbox.utils import singleton


def test_singleton():
    @singleton
    class A(object):
        def __init__(self, *args, **kwargs):
            pass

    obj1 = A(name='Siddhesh')
    obj2 = A(name='Siddhesh', lname='Sathe')
    obj3 = A(name='Siddhesh', lname='Sathe')
    assert obj1 is not obj2, "Created objects must be different"
    assert obj2 is obj3, "Created objects must be same"




import pytest
import yaml
import asynctest
import aiohttp

from unittest.mock import patch, MagicMock

from .fixtures import SessionMock, ResponseMock
from pathlib import Path

from aionetbox.api import (
    NetboxResponseObject,
    NetboxApiOperation,
    NetboxApi,
    AIONetbox,
    ResolvingParser,
)

from aionetbox.exceptions import (
    InvalidResponse,
    MissingRequiredParam,
    BadRequest,
    InvalidRequest,
)

here = Path(__file__).parent


def test_NetboxResponseObject_from_response_no_type():

    with pytest.raises(AttributeError):
        NetboxResponseObject.from_response(data={})


def test_NetboxResponseObject_from_response():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))

    schema = spec.specification['paths']['/users/{userId}']['get']['responses']['200']['schema']

    data = {
        'id': 1,
        'name': 'Marco Ceppi',
        'cars': [
            {
                'make': 'MG',
                'model': 'BGT',
                'year': 1974,
                'problems': None
            },
            {
                'make': 'Triumph',
                'model': 'TR6',
                'year': 1974,
                'problems': [
                    'oil leak'
                ]
            }
        ],
        'address': {
            'number': 123,
            'street': 'Main St',
        }
    }

    n = NetboxResponseObject.from_response(data=data, **schema)

    assert n.id == 1
    assert n.cars[0].model == 'BGT'
    assert n['cars'][0]['model'] == 'BGT'
    assert len(n) == 4

    with pytest.raises(KeyError):
        n['notvalid']

    for k, val in n:
        assert n[k] == val


def test_NetboxResponseObject_repr():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    schema = spec.specification['paths']['/users/{userId}']['get']['responses']['200']['schema']

    data = {
        'id': 1,
        'name': 'Marco Ceppi',
        'address': {
            'number': 123,
            'street': 'Main St',
        }
    }

    n = NetboxResponseObject.from_response(data=data, **schema)

    assert yaml.safe_load(str(n)) == data


def test_NetboxResponseObject_from_invalid_response():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    schema = spec.specification['paths']['/users/{userId}']['get']['responses']['200']['schema']

    data = {
        'id': 1,
        'name': 'Marco Ceppi',
        'cars': [
            {
                'make': 'MG',
                'model': 'BGT',
                'year': 1974,
            },
            {
                'model': 'TR6',
                'year': 1974,
            }
        ]
    }

    with pytest.raises(InvalidResponse):
        NetboxResponseObject.from_response(data=data, **schema)


def test_NetboxApiOperation():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    op = NetboxApiOperation('users', 'users_read', {}, nb)

    assert repr(op) == 'Netbox.users.users_read'


def test_NetboxApiOperation_build_url():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    op = NetboxApiOperation('users', 'users_read', {}, nb)

    assert 'http://localhost/v1/test' == op.build_url('/test')


def test_NetboxApiOperation_parse_params():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    op = NetboxApiOperation('users', 'users_list', nb.config['users']['users_list'], nb)

    assert ({}, {}, {}) == op.parse_params({})

    params = {'name': 'foobar', 'bad_param': True}
    parsed = op.parse_params(params)

    assert ({}, {}, {'name': 'foobar'}) == parsed

    op = NetboxApiOperation('users', 'users_read', nb.config['users']['users_read'], nb)

    params = {'userId': 1}
    parsed = op.parse_params(params)

    assert ({'userId': 1}, {}, {}) == parsed

    with pytest.raises(MissingRequiredParam):
        op.parse_params({})


def test_NetboxApiOperation_parse_params_body():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    op = NetboxApiOperation('users', 'users_create', nb.config['users']['users_create'], nb)

    params = {'data': {'foobar': 'bar', 'baz': True}}
    parsed = op.parse_params(params)

    assert ({}, {'foobar': 'bar', 'baz': True}, {}) == parsed


def test_NetboxApiOperation_operation_method():
    op = NetboxApiOperation('users', 'users_read', {}, {})
    assert 'read' == op.operation_method

    op = NetboxApiOperation('users', 'users_partial_update', {}, {})
    assert 'partial_update' == op.operation_method

    op = NetboxApiOperation('users', 'users_update', {}, {})
    assert 'update' == op.operation_method


@pytest.mark.asyncio
async def test_NetboxApiOperation_call():

    with patch.object(NetboxApiOperation, 'request') as mreq:
        mreq.side_effect = asynctest.MagicMock()
        op = NetboxApiOperation('users', 'users_read', {}, {})
        await op(foo='test')
        mreq.assert_called_with(foo='test')

    with patch.object(NetboxApiOperation, 'request') as mreq:
        mreq.side_effect = [MissingRequiredParam, aiohttp.ClientResponseError({}, {})]
        op = NetboxApiOperation('users', 'users_read', {}, {})

        with pytest.raises(AttributeError):
            await op(foo='test')

        with pytest.raises(BadRequest):
            await op(foo='test')


@pytest.mark.asyncio
@patch('aionetbox.api.NetboxResponseObject')
async def test_NetboxApiOperation_request(mnbro):
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    nb.request = asynctest.CoroutineMock(return_value=ResponseMock())
    op = NetboxApiOperation('users', 'users_read', nb.config['users']['users_read'], nb)

    response_schema = spec.specification['paths']['/users/{userId}']['get']['responses']['200']['schema']
    op.parse_params = MagicMock(return_value=({}, {}, {}))
    op.build_url = MagicMock()

    await op.request(userId=42)

    mnbro.from_response.assert_called_with(data={}, **response_schema)


@pytest.mark.asyncio
@patch('aionetbox.api.NetboxResponseObject')
async def test_NetboxApiOperation_request_cf(mnbro):
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    nb.request = asynctest.CoroutineMock(return_value=ResponseMock())
    op = NetboxApiOperation('users', 'users_list', nb.config['users']['users_list'], nb)

    mnbro.from_response.return_value.next = None

    response_schema = spec.specification['paths']['/users']['get']['responses']['200']['schema']
    op.parse_params = MagicMock(return_value=({}, {}, {'name': 'test'}))
    op.build_url = MagicMock(return_value='http://test/users')

    await op.request(name='test', cf_hello='world', extra='ignore')

    nb.request.assert_called_with(
        method='get',
        url='http://test/users',
        query_params={'name': 'test', 'cf_hello': 'world'},
        body={}
    )

    mnbro.from_response.assert_called_with(data={}, **response_schema)


@pytest.mark.asyncio
@patch('aionetbox.api.NetboxResponseObject')
async def test_NetboxApiOperation_request_pagination(mnbro):
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    nb.request = asynctest.CoroutineMock(return_value=ResponseMock())
    op = NetboxApiOperation('users', 'users_list', nb.config['users']['users_list'], nb)

    page0 = MagicMock()
    page1 = MagicMock()

    page0.next = 'http://testing/test?limit=50&offset=50'
    page0.results = ['1', '2', '3']
    page1.next = None
    page1.results = ['one', 'two', 'three']

    mnbro.from_response.side_effect = [page0, page1]

    op.parse_params = MagicMock(return_value=({}, {}, {'name': 'test'}))
    op.build_url = MagicMock(return_value='http://test/users')

    results = await op.request(name='test')

    nb.request.assert_any_call(
        method='get',
        url='http://test/users',
        query_params={'name': 'test'},
        body={}
    )

    nb.request.assert_any_call(
        method='get',
        url='http://testing/test?limit=50&offset=50'
    )

    assert ['1', '2', '3', 'one', 'two', 'three'] == results.results


def test_NetboxApi():
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())
    api = NetboxApi('users', nb.config.get('users'), nb)

    assert repr(api) == 'Netbox.users'


@patch('aionetbox.api.NetboxApiOperation')
def test_NetboxApi_attr(mnbaop):
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())

    api = NetboxApi('users', nb.config.get('users'), nb)

    api.users_read
    api.users_read

    mnbaop.assert_called_once()
    mnbaop.assert_called_with('users', 'users_read', nb.config['users']['users_read'], nb)

    with pytest.raises(AttributeError):
        api.fake_method


@patch('aionetbox.api.NetboxSpec')
def test_AIONetbox_from_openapi(mrp):
    with patch.object(AIONetbox, 'parse_spec') as mps:
        AIONetbox.from_openapi('http://testurl', 'ab12cd34')
        mrp.assert_called_with('http://testurl/api/swagger.json')
        mps.assert_called_with(mrp.return_value)


def test_AIONetbox_from_spec():
    a = AIONetbox.from_spec(str(here / 'data' / 'openapi-2.yaml'), 'ab12cd34')

    assert a.host == 'https://api.example.com'


def test_AIONetbox_parse_spec():
    a = AIONetbox('http://localhost', '1122')
    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))

    assert {} == a.parse_spec(None)
    assert {} == a.parse_spec({})
    assert '_orig' in a.parse_spec(spec)
    assert 'users' in a.parse_spec(spec)
    assert 'fake' not in a.parse_spec(spec)
    assert 'users_list' in a.parse_spec(spec).get('users')


@pytest.mark.asyncio
async def test_AIONetbox_request_bad_params():
    s = SessionMock()
    a = AIONetbox('http://localhost', '1122', session=s)

    with pytest.raises(ValueError):
        await a.request('get', 'http://foobar', post_params={'hello': 'world'}, body={'wrong': True})

    s.request.assert_not_called()


@pytest.mark.asyncio
async def test_AIONetbox_request_invalid_content_type():
    s = SessionMock()
    a = AIONetbox('http://localhost', '1122', session=s)

    with pytest.raises(InvalidRequest):
        await a.request('put', 'http://foobar', headers={'Content-Type': 'broken'})

    s.request.assert_not_called()


@pytest.mark.asyncio
async def test_AIONetbox_request():
    s = SessionMock()
    a = AIONetbox('http://localhost', '1122', session=s)
    await a.request('post', 'http://foobar', query_params={'test': 'testing'}, body={'new_phone': 'whodis'})

    s.request.assert_called_with(
        method='POST',
        url='http://foobar',
        timeout=300,
        params={'test': 'testing'},
        headers={
            'Content-Type': 'application/json',
            'Authorization': 'Token 1122',
        },
        data='{"new_phone": "whodis"}'
    )


@pytest.mark.asyncio
@patch('aionetbox.api.aiohttp.FormData')
async def test_AIONetbox_request_form(maiofd):
    s = SessionMock()
    a = AIONetbox('http://localhost', '1122', session=s)

    await a.request(
        'post',
        'http://foobar',
        headers={'Content-Type': 'application/x-www-form-urlencoded'},
        post_params={'new_phone': 'whodis'}
    )

    maiofd.assert_called_with({'new_phone': 'whodis'})

    s.request.assert_called_with(
        method='POST',
        url='http://foobar',
        timeout=300,
        params=None,
        headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Token 1122',
        },
        data=maiofd.return_value
    )


@pytest.mark.asyncio
async def test_AIONetbox_request_bytes():
    s = SessionMock()
    a = AIONetbox('http://localhost', '1122', session=s)

    data = b'yaml: cool'
    await a.request(
        'post',
        'http://foobar',
        headers={'Content-Type': 'application/yaml'},
        body=data
    )

    s.request.assert_called_with(
        method='POST',
        url='http://foobar',
        timeout=300,
        params=None,
        headers={
            'Content-Type': 'application/yaml',
            'Authorization': 'Token 1122',
        },
        data=data
    )


@pytest.mark.asyncio
async def test_AIONetbox_close():
    s = SessionMock()
    nb = AIONetbox(host='http://localhost', api_key='11', spec={}, session=s)
    await nb.close()
    s.close.assert_called()


@patch('aionetbox.api.NetboxApi')
def test_AIONetbox_attr(mnba):

    spec = ResolvingParser(str(here / 'data' / 'openapi-2.yaml'))
    nb = AIONetbox(host='http://localhost', api_key='11', spec=spec, session=SessionMock())

    nb.users
    nb.users

    mnba.assert_called_once()
    mnba.assert_called_with('users', nb.config['users'], nb)

    with pytest.raises(AttributeError):
        nb.fake_tag




import re
import json
import asyncio
import logging
import collections

from aionetbox.utils import singleton

from typing import Iterable

import aiohttp
from prance import ResolvingParser

from aionetbox.exceptions import (
    BadRequest,
    MissingRequiredParam,
    InvalidResponse,
    InvalidRequest,
)

log = logging.getLogger(__name__)


@singleton
class NetboxSpec(ResolvingParser):
    pass


class AIONetbox:
    """asyncio Netbox Client

    Args:
        host: Base URL of Netbox instance
        api_key: The API Token from Netbox to access Netbox API
        spec: Dictionary or ResolvingParser of OpenAPI/Swagger spec
        session: ClientSession for http requests
        loop: Optional asyncio loop
    """

    _http_methods = ('GET', 'HEAD', 'DELETE', 'POST', 'PUT', 'PATCH', 'OPTIONS')
    _api_cache = {}

    @classmethod
    def from_openapi(cls, url, api_key, session=None):
        """Create an AIONetbox using a remote OpenAPI/Swagger Specification.

        Args:
            url: Base URL with protocol to load swagger spec.
            api_key: The API Token from Netbox to access Netbox API
            session: ClientSession for http requests
        """
        spec = NetboxSpec('{}/api/swagger.json'.format(url))

        aionb = cls(url, api_key, spec=spec, session=session)

        return aionb

    @classmethod
    def from_spec(cls, spec, api_key, session=None):
        """Create an AIONetbox using an OpenAPI/Swagger Specification from disk.

        Args:
            spec: String or full path on disk of OpenAPI/Swagger spec
            api_key: The API Token from Netbox to access Netbox API
            session: ClientSession for http requests
        """
        data = NetboxSpec(spec)

        url = '{}://{}'.format(data.specification.get('schemes', ['http'])[0], data.specification.get('host'))
        aionb = cls(url, api_key, spec=data, session=session)

        return aionb

    def __init__(self, host, api_key, spec=None, session=None, loop=None):
        self.host = host
        self.api_key = api_key
        self.session = session or aiohttp.ClientSession()
        self.config = self.parse_spec(spec)
        self.tags = self.config.keys()
        self.loop = loop or asyncio.get_event_loop()

    async def request(self, method, url, query_params=None, headers=None, body=None, post_params=None, timeout=None):
        """Execute a web request

        Args:
            method: HTTP verb for request
            url: Full URL of resource for request
            query_params: Query paramemters to be added at the end of url
            headers: Additional headers for request
            body: Request body
            post_params: http form parameters
            timeout: request timeout in seconds
        """
        method = method.upper()
        assert method in self._http_methods

        if post_params and body:
            raise ValueError(
                'body parameter cannot be used with post_params parameter.'
            )

        post_params = post_params or {}
        headers = headers or {}
        headers.update({
            'Authorization': f'Token {self.api_key}',
        })
        timeout = timeout or 5 * 60

        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'

        args = {
            'method': method,
            'url': url,
            'timeout': timeout,
            'params': query_params,
            'headers': headers
        }

        if method in ['POST', 'PUT', 'PATCH', 'OPTIONS', 'DELETE']:
            if re.search('json', headers['Content-Type'], re.IGNORECASE):
                if body is not None:
                    body = json.dumps(body)
                args['data'] = body
            elif headers['Content-Type'] == 'application/x-www-form-urlencoded':  # noqa: E501
                args['data'] = aiohttp.FormData(post_params)

            # Pass a `bytes` parameter directly in the body to support
            # other content types than Json when `body` argument is provided
            # in serialized form
            elif isinstance(body, bytes):
                args['data'] = body
            else:
                # Cannot generate the request from given parameters
                raise InvalidRequest(
                    'Cannot prepare a request message for provided arguments. Please check that your arguments match '
                    'declared content type.'
                )

        return await self.session.request(**args)

    def parse_spec(self, config):
        """Parse OpenAPI/Swagger spec"""

        operations = {}

        if not config:
            return {}

        if isinstance(config, ResolvingParser):
            config = config.specification

        operations['_orig'] = config
        for path, data in config.get('paths', {}).items():
            for method in self._http_methods:
                action = method.lower()
                if action not in data:
                    continue

                payload = data.get(action, {})
                tags = payload.get('tags', [])

                operationId = payload.get('operationId')
                if not operationId:
                    continue

                for tag in tags:
                    if tag not in operations:
                        operations[tag] = {}

                    rest_cfg = {
                        'url': path,
                        'params': data.get('parameters', []),
                        'method': action,
                    }

                    operations[tag][operationId] = payload
                    operations[tag][operationId]['rest'] = rest_cfg

        return operations

    def __getattr__(self, tag):
        """NetboxApi Lookup

        Used when accessing a certain tag of the netbox api. The result of this method depends on what tags
        are defined in the OpenAPI/Swagger spec.
        """
        if tag not in self.tags:
            raise AttributeError("'{}' object has no attribute '{}'".format(self.__class__.__name__, tag))

        if tag in self._api_cache:
            return self._api_cache.get(tag)

        c = NetboxApi(tag, self.config.get(tag), self)
        self._api_cache[tag] = c

        return c

    async def close(self):
        """Close all open connections"""
        await self.session.close()

    def __del__(self):
        asyncio.ensure_future(self.session.close())


class NetboxApi:
    """Netbox API group endpoint

    Args:
        tag: Netbox API Group name
        operations: one or more operations
        client: AIONetbox instance
    """
    def __init__(self, tag, operations, client):
        self.name = tag
        self.config = operations
        self.operations = operations.keys()
        self.client = client

        self._operation_cache = {}

    def __getattr__(self, operation):
        """NetboxApiOperation lookup

        Used to load an operationId from configuration of the API group
        """
        if operation not in self.operations:
            raise AttributeError("'{}' object has no attribute '{}'".format(self, operation))

        if operation in self._operation_cache:
            return self._operation_cache.get(operation)

        m = NetboxApiOperation(self.name, operation, self.config.get(operation), self.client)
        self._operation_cache[operation] = m

        return m

    def __repr__(self):
        return 'Netbox.{}'.format(self.name)


class NetboxApiOperation:
    """Netbox API group operation

    Args:
        tag: Netbox API Group name
        operation: operation name
        config: Operation configuration
        client: AIONetbox instance
    """
    def __init__(self, tag, operation, config, client):
        self.client = client
        self.config = config
        self.rest_config = self.config.get('rest', {})
        self.operation = operation
        self.tag = tag

    async def request(self, **kwargs):
        """Execute a web request"""
        path, body, query = self.parse_params(kwargs)

        if self.operation_method == 'list':
            # netbox has terrible swagger spec generation and leaves out custom fields (cf)
            keys_used = set(list(body.keys()) + list(query.keys()) + list(path.keys()))
            remaining_vars = set(kwargs.keys()) - keys_used

            for kw in remaining_vars:
                if kw.startswith('cf_'):
                    query[kw] = kwargs.get(kw)

        url = self.build_url(self.rest_config.get('url')).format(**path)

        resp = await self.client.request(
            method=self.rest_config.get('method'),
            url=url,
            query_params=query,
            body=body
        )

        resp.raise_for_status()
        data = await resp.json()
        output = NetboxResponseObject.from_response(
            data=data,
            **self.config.get('responses', {}).get(str(resp.status), {}).get('schema', {})
        )

        if self.operation_method != 'list':
            return output

        while output.next:
            resp = await self.client.request(
                method=self.rest_config.get('method'),
                url=output.next
            )

            resp.raise_for_status()
            data = await resp.json()
            pagination_output = NetboxResponseObject.from_response(
                data=data,
                **self.config.get('responses', {}).get(str(resp.status), {}).get('schema', {})
            )

            output.results.extend(pagination_output.results)
            output.next = pagination_output.next

        return output

    def build_url(self, url):
        """Construct a URL from spec"""
        return '{}{}{}'.format(self.client.host, self.client.config.get('_orig', {}).get('basePath'), url)

    def parse_params(self, params):
        """Build request parameters and validate based on spec"""
        qb = {
            'body': {},
            'query': {},
            'path': {},
        }

        spec_parameters = self.config.get('parameters', []) + self.rest_config.get('params', [])
        for sp in spec_parameters:
            if sp['name'] not in params and sp['required']:
                raise MissingRequiredParam('{} is a required parameter'.format(sp['name']))

            if sp['name'] not in params:
                continue

            if sp['in'] == 'body':
                # BODY params are special in that the name is not passed through
                qb[sp['in']] = params.get(sp['name'])
            else:
                qb[sp['in']][sp['name']] = params.get(sp['name'])

        return (qb.get('path'), qb.get('body'), qb.get('query'))

    @property
    def operation_method(self):
        if self.operation.endswith('partial_update'):
            return 'partial_update'

        return self.operation.split('_')[-1]

    async def __call__(self, **kwargs):
        """ Execute given operation request

        When a user executes an instance of NetboxApiOperation those arguments will passed through to ``request``
        """
        try:
            return await self.request(**kwargs)
        except MissingRequiredParam as e:
            raise AttributeError(str(e))
        except aiohttp.ClientResponseError as e:
            raise BadRequest('{} failed with code {}: {}'.format(self.operation, e.status, e.message))

    def __repr__(self):
        return 'Netbox.{}.{}'.format(self.tag, self.operation)


class NetboxResponseObject:
    """Netbox Response

    Take the output from an ``AIONetbox.request`` and formats / validates it based on a passed spec
    """

    @classmethod
    def from_response(cls, data, **kwargs):

        if 'type' not in kwargs:
            raise AttributeError('type is a required parameter')

        output = cls()
        if kwargs['type'] != 'object':
            return data

        required = kwargs.get('required', [])
        properties = kwargs.get('properties', {})

        for req in required:
            # Check for all required response parameters
            if req not in data:
                raise InvalidResponse('Response did not include required "{}"'.format(req))

        # if type == 'array':
        #     return [] # What if the root object is an array?

        # Iterate over every datapoint and further parse if required
        for key, val in data.items():
            spec = properties.get(key, {})
            ctype = spec.get('type')
            value = val

            if ctype == 'object':
                if isinstance(val, collections.Mapping):
                    value = cls.from_response(data=val, **spec)
            elif ctype == 'array':
                # If we have an array of objects, make sure the value is iterable, then produce a list of objects
                if isinstance(val, Iterable):
                    value = [cls.from_response(data=v, **spec.get('items')) for v in val]

            setattr(output, key, value)

        return output

    def __iter__(self):
        iters = {}
        iters.update(self.__dict__)

        for k, val in iters.items():
            yield k, val

    def __getitem__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        raise KeyError(key)

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        return repr(self.__dict__)




from .api import AIONetbox

__title__ = 'aionetbox'
__version__ = '1.3.0'
__description__ = 'Async netbox library'
__author__ = 'Marco Ceppi'
__author_email__ = 'marco@vapor.io'
__url__ = 'https://github.com/vapor-ware/aionetbox'
__license__ = 'GNU General Public License v3.0'



#!/usr/bin/env python

import os
import re

from pathlib import Path
from typing import List

from codecs import open
from setuptools import find_packages, setup


here = Path(__file__).parent


# Load the package's __init__.py file as a dictionary.
pkg = {}
with open(here / 'aionetbox' / '__init__.py', 'r', 'utf-8') as f:
    pkg = {k: v for k, v in re.findall(r"^(__\w+__) = \'(.+)\'", f.read(), re.M)}

# Load the README
readme = ''
if os.path.exists(here / 'README.md'):
    with open(here / 'README.md', 'r', 'utf-8') as f:
        readme = f.read()

setup(
    name=pkg['__title__'],
    version=pkg['__version__'],
    description=pkg['__description__'],
    license=pkg['__license__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    url=pkg['__url__'],
    author=pkg['__author__'],
    author_email=pkg['__author_email__'],
    packages=find_packages(),
    package_data={
        '': ['LICENSE']
    },
    package_dir={
        'aionetbox': 'aionetbox',
    },
    python_requires='>=3.6',
    install_requires=[
        'aiohttp[speedups]',
        'prance[osv]',
    ],
    extras_require={
        'icu': [
            'prance[osv,icu]',
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    zip_safe=False,
)
